==> tfx-1.5.1-py3-none-any.whl: tfx/examples/chicago_taxi_pipeline/serving/chicago_taxi_client.py <==
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A client for the chicago_taxi demo."""

import argparse
import base64
import csv
import json
import os
import subprocess
import tempfile
from typing import List

from absl import app
from absl.flags import argparse_flags
import requests
from tensorflow_transform import coders as tft_coders
from tensorflow_transform.tf_metadata import schema_utils

from google.protobuf import text_format
from tensorflow.python.lib.io import file_io  # pylint: disable=g-direct-tensorflow-import
from tensorflow_metadata.proto.v0 import schema_pb2

_LOCAL_INFERENCE_TIMEOUT_SECONDS = 5.0

_LABEL_KEY = 'tips'


# Tf.Transform considers these features as "raw"
def _get_raw_feature_spec(schema):
  return schema_utils.schema_as_feature_spec(schema).feature_spec


def _make_proto_coder(schema):
  raw_feature_spec = _get_raw_feature_spec(schema)
  raw_schema = schema_utils.schema_from_feature_spec(raw_feature_spec)
  return tft_coders.ExampleProtoCoder(raw_schema)


def _read_schema(path):
  """Reads a schema from the provided location.

  Args:
    path: The location of the file holding a serialized Schema proto.

  Returns:
    An instance of Schema or None if the input argument is None
  """
  result = schema_pb2.Schema()
  contents = file_io.read_file_to_string(path)
  text_format.Parse(contents, result)
  return result


def _do_local_inference(host, port, serialized_examples):
  """Performs inference on a model hosted by the host:port server."""

  json_examples = []
  for serialized_example in serialized_examples:
    # The encoding follows the guidelines in:
    # https://www.tensorflow.org/tfx/serving/api_rest
    example_bytes = base64.b64encode(serialized_example).decode('utf-8')
    predict_request = '{ "b64": "%s" }' % example_bytes
    json_examples.append(predict_request)

  json_request = '{ "instances": [' + ','.join(map(str, json_examples)) + ']}'

  server_url = 'http://' + host + ':' + port + '/v1/models/chicago_taxi:predict'
  response = requests.post(
      server_url, data=json_request, timeout=_LOCAL_INFERENCE_TIMEOUT_SECONDS)
  response.raise_for_status()
  prediction = response.json()
  print(json.dumps(prediction, indent=4))


def _do_aiplatform_inference(model, version, serialized_examples):
  """Performs inference on the model:version in AI Platform."""
  working_dir = tempfile.mkdtemp()
  instances_file = os.path.join(working_dir, 'test.json')
  json_examples = []
  for serialized_example in serialized_examples:
    # The encoding follows the example in:
    # https://github.com/GoogleCloudPlatform/training-data-analyst/blob/master/quests/tpu/invoke_model.py
    json_examples.append('{ "inputs": { "b64": "%s" } }' %
                         base64.b64encode(serialized_example).decode('utf-8'))
  file_io.write_string_to_file(instances_file, '\n'.join(json_examples))
  gcloud_command = [
      'gcloud', 'ai-platform', 'predict', '--model', model, '--version',
      version, '--json-instances', instances_file
  ]
  print(subprocess.check_output(gcloud_command))


def _do_inference(model_handle, examples_file, num_examples, schema):
  """Sends requests to the model and prints the results.

  Args:
    model_handle: handle to the model. This can be either
     "aiplatform:model:version" or "host:port"
    examples_file: path to csv file containing examples, with the first line
      assumed to have the column headers
    num_examples: number of requests to send to the server
    schema: a Schema describing the input data

  Returns:
    Response from model server
  """
  filtered_features = [
      feature for feature in schema.feature if feature.name != _LABEL_KEY
  ]
  del schema.feature[:]
  schema.feature.extend(filtered_features)

  proto_coder = _make_proto_coder(schema)

  csv_reader = csv.DictReader(open(examples_file, 'r'))

  serialized_examples = []
  for _ in range(num_examples):
    one_line = next(csv_reader)
    if not one_line:
      print('End of example file reached')
      break
    one_example = {}
    for feature in schema.feature:
      name = feature.name
      if one_line[name]:
        if feature.type == schema_pb2.FLOAT:
          one_example[name] = [float(one_line[name])]
        elif feature.type == schema_pb2.INT:
          one_example[name] = [int(one_line[name])]
        elif feature.type == schema_pb2.BYTES:
          one_example[name] = [one_line[name].encode('utf8')]
      else:
        one_example[name] = []

    serialized_example = proto_coder.encode(one_example)
    serialized_examples.append(serialized_example)

  parsed_model_handle = model_handle.split(':')
  if parsed_model_handle[0] == 'aiplatform':
    _do_aiplatform_inference(
        model=parsed_model_handle[1],
        version=parsed_model_handle[2],
        serialized_examples=serialized_examples)
  else:
    _do_local_inference(
        host=parsed_model_handle[0],
        port=parsed_model_handle[1],
        serialized_examples=serialized_examples)


def _parse_flags(argv: List[str]) -> argparse.Namespace:
  """Command lines flag parsing."""
  parser = argparse_flags.ArgumentParser()
  parser.add_argument(
      '--num_examples',
      help=('Number of examples to send to the server.'),
      default=1,
      type=int)

  parser.add_argument(
      '--server',
      help=('Prediction service host:port or aiplatform:model:version'),
      required=True)

  parser.add_argument(
      '--examples_file',
      help=('Path to csv file containing examples.'),
      required=True)

  parser.add_argument(
      '--schema_file', help='File holding the schema for the input data')
  return parser.parse_args(argv)


def main(args: argparse.Namespace):
  _do_inference(args.server, args.examples_file,
                args.num_examples, _read_schema(args.schema_file))


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)

==> tfx-1.13.0-py3-none-any.whl: tfx/examples/chicago_taxi_pipeline/serving/chicago_taxi_client.py <==
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A client for the chicago_taxi demo."""

import argparse
import base64
import csv
import json
import os
import subprocess
import tempfile
from typing import List

from absl import app
from absl.flags import argparse_flags
import requests
import tensorflow_data_validation as tfdv
from tensorflow_transform import coders as tft_coders
from tensorflow_transform.tf_metadata import schema_utils

from google.protobuf import text_format
from tensorflow.python.lib.io import file_io  # pylint: disable=g-direct-tensorflow-import
from tensorflow_metadata.proto.v0 import schema_pb2

_LOCAL_INFERENCE_TIMEOUT_SECONDS = 5.0

_LABEL_KEY = 'tips'


# Tf.Transform considers these features as "raw"
def _get_raw_feature_spec(schema):
  return schema_utils.schema_as_feature_spec(schema).feature_spec


def _make_proto_coder(schema):
  raw_feature_spec = _get_raw_feature_spec(schema)
  raw_schema = schema_utils.schema_from_feature_spec(raw_feature_spec)
  return tft_coders.ExampleProtoCoder(raw_schema)


def _read_schema(path):
  """Reads a schema from the provided location.

  Args:
    path: The location of the file holding a serialized Schema proto.

  Returns:
    An instance of Schema or None if the input argument is None
  """
  result = schema_pb2.Schema()
  contents = file_io.read_file_to_string(path)
  text_format.Parse(contents, result)
  return result


def _do_local_inference(host, port, serialized_examples):
  """Performs inference on a model hosted by the host:port server."""

  json_examples = []
  for serialized_example in serialized_examples:
    # The encoding follows the guidelines in:
    # https://www.tensorflow.org/tfx/serving/api_rest
    example_bytes = base64.b64encode(serialized_example).decode('utf-8')
    predict_request = '{ "b64": "%s" }' % example_bytes
    json_examples.append(predict_request)

  json_request = '{ "instances": [' + ','.join(map(str, json_examples)) + ']}'

  server_url = 'http://' + host + ':' + port + '/v1/models/chicago_taxi:predict'
  response = requests.post(
      server_url, data=json_request, timeout=_LOCAL_INFERENCE_TIMEOUT_SECONDS)
  response.raise_for_status()
  prediction = response.json()
  print(json.dumps(prediction, indent=4))


def _do_aiplatform_inference(model, version, serialized_examples):
  """Performs inference on the model:version in AI Platform."""
  working_dir = tempfile.mkdtemp()
  instances_file = os.path.join(working_dir, 'test.json')
  json_examples = []
  for serialized_example in serialized_examples:
    # The encoding follows the example in:
    # https://github.com/GoogleCloudPlatform/training-data-analyst/blob/master/quests/tpu/invoke_model.py
    json_examples.append('{ "inputs": { "b64": "%s" } }' %
                         base64.b64encode(serialized_example).decode('utf-8'))
  file_io.write_string_to_file(instances_file, '\n'.join(json_examples))
  gcloud_command = [
      'gcloud', 'ai-platform', 'predict', '--model', model, '--version',
      version, '--json-instances', instances_file
  ]
  print(subprocess.check_output(gcloud_command))


def _do_inference(model_handle, examples_file, num_examples, schema):
  """Sends requests to the model and prints the results.

  Args:
    model_handle: handle to the model. This can be either
     "aiplatform:model:version" or "host:port"
    examples_file: path to csv file containing examples, with the first line
      assumed to have the column headers
    num_examples: number of requests to send to the server
    schema: a Schema describing the input data

  Returns:
    Response from model server
  """
  filtered_features = [
      feature for feature in schema.feature if feature.name != _LABEL_KEY
  ]
  del schema.feature[:]
  schema.feature.extend(filtered_features)

  proto_coder = _make_proto_coder(schema)

  csv_reader = csv.DictReader(open(examples_file, 'r'))

  dataset_stats = tfdv.generate_statistics_from_csv(examples_file)
  serialized_examples = []
  for _ in range(num_examples):
    one_line = next(csv_reader)
    if not one_line:
      print('End of example file reached')
      break
    one_example = {}
    for feature in schema.feature:
      name = feature.name
      feature_stats = tfdv.get_feature_stats(dataset_stats.datasets[0],
                                             tfdv.FeaturePath([name]))
      if one_line[name]:
        if feature.type == schema_pb2.FLOAT:
          one_example[name] = [float(one_line[name])]
        elif feature.type == schema_pb2.INT:
          one_example[name] = [int(one_line[name])]
        elif feature.type == schema_pb2.BYTES:
          one_example[name] = [one_line[name].encode('utf8')]
      else:
        # TF serve does not like missing features, so we'll populate
        # the missing features with their mean/mode instead
        if feature.type == schema_pb2.FLOAT:
          one_example[name] = [feature_stats.num_stats.mean]
        elif feature.type == schema_pb2.INT:
          one_example[name] = [int(feature_stats.num_stats.mean)]
        elif feature.type == schema_pb2.BYTES:
          top_values = list(feature_stats.string_stats.top_values)
          if top_values:
            one_example[name] = [top_values[0].value.encode('utf8')]
          else:
            one_example[name] = [''.encode('utf8')]

    serialized_example = proto_coder.encode(one_example)
    serialized_examples.append(serialized_example)

  parsed_model_handle = model_handle.split(':')
  if parsed_model_handle[0] == 'aiplatform':
    _do_aiplatform_inference(
        model=parsed_model_handle[1],
        version=parsed_model_handle[2],
        serialized_examples=serialized_examples)
  else:
    _do_local_inference(
        host=parsed_model_handle[0],
        port=parsed_model_handle[1],
        serialized_examples=serialized_examples)


def _parse_flags(argv: List[str]) -> argparse.Namespace:
  """Command lines flag parsing."""
  parser = argparse_flags.ArgumentParser()
  parser.add_argument(
      '--num_examples',
      help=('Number of examples to send to the server.'),
      default=1,
      type=int)

  parser.add_argument(
      '--server',
      help=('Prediction service host:port or aiplatform:model:version'),
      required=True)

  parser.add_argument(
      '--examples_file',
      help=('Path to csv file containing examples.'),
      required=True)

  parser.add_argument(
      '--schema_file', help='File holding the schema for the input data')
  return parser.parse_args(argv[1:])


def main(args: argparse.Namespace):
  _do_inference(args.server, args.examples_file,
                args.num_examples, _read_schema(args.schema_file))


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)
